/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.pdfbox.pdfparser.tokens;

import java.io.IOException;

import org.apache.pdfbox.io.RandomAccessRead;


/**
 * The PDFLexer allows forward, read-only access to individual tokens within a
 * binary representation of a PDF. It is designed to be the a low level way to
 * read PDF data. These tokens will be further processed by a PDF parsing
 * component.
 * 
 * <p>
 * The PDFLexer is designed to iterate over PDF using next() and hasNext(). The
 * data can be accessed using methods such as getEventType() and getToken().
 * </p>
 * 
 * <p>
 * The <a href="#next()">next()</a> method causes the lexer to read the next
 * parse event. The next() method returns a PDF parsing event which identifies
 * the type of event just read.
 * </p>
 * 
 * <p>
 * The event type can be determined using <a href="#getEventType()">
 * getEventType()</a>. The list of possible event types is defined in
 * {@link TokenTypes}
 * </p>
 * 
 * <p>The events represent most of the basic PDF Objects as specified in
 * ISO-32000-1:2008 7.2. and 7.3. These are
 * </p>
 * 
 * <ul>
 * <li>EOL [IS0 32000-1:2008: 7.2.2 Tab 1]</li>
 * <li>Whitespace [IS0 32000-1:2008: 7.2.2 Tab 1]</li>
 * <li>Comment [IS0 32000-1:2008: 7.2.3]</li>
 * <li>Boolean Objects [IS0 32000-1:2008: 7.3.2]</li>
 * <li>Numeric Objects [IS0 32000-1:2008: 7.3.3]</li>
 * <li>String Objects [IS0 32000-1:2008: 7.3.4]</li>
 * <li>Name Object [IS0 32000-1:2008: 7.3.5]</li>
 * <li>Keyword [IS0 32000-1:2008: 7.3]</li>
 * <li>Array Objects [IS0 32000-1:2008: 7.3.6]</li>
 * <li>Dictionary Objects [IS0 32000-1:2008: 7.3.7]</li>
 * <li>Stream Objects [IS0 32000-1:2008: 7.3.8]</li>
 * <li>Null Objects [IS0 32000-1:2008: 7.3.9]</li>
 * <li>Unknown</li>
 * </ul>
 * 
 * <p>In addition events are delivered for:</>
 * 
 * <ul>
 * <li>start and end of document</li>
 * <li>start and end of a dictionary object</li>
 * <li>start and end of a array object<li>
 * </ul>
 * 
 * <p>Indirect Objects need to be build based on the basic event types.</p>
 * 
 * <p>
 * The data within the tokens can be requested in it's raw format using
 * {@link #getToken()}. It's up to an application using this class to check for
 * the validity of the data passed e.g. check if literal strings are terminated
 * correctly. 
 * </p>
 * 
 * <p>
 * Reading the content of the tokens is deferred to calling {@link #getToken()}.
 * It's up to the application to persist the tokens data.
 * </p>
 * 
 * 
 * <p>
 * About the implementation
 * </p>
 * <p>
 * The lexer is reading one byte a time and inspects that for certain delimiters
 * or whitespace characters. Dependent on the character another byte might be
 * read as some basic PDF objects are marked by two char delimiters.
 * </p>
 * 
 * <p>For keywords the data is processed until the end of the keyword to be able
 * to gather the necessary information for boolean objects, stream objects and
 * null objects</p>
 * 
 * <p>
 * Characters which are not part of the current event are put back into a simple
 * buffer.
 * </p>
 * 
 * @author Maruan Sahyoun (sahyoun at fileaffairs.de)
 * @version $Revision: 1.0 $
 */
public class PDFLexer
{

    /**
     * Definition of the list of whitespace characters.
     * [IS0 32000-1:2008: 7.2.2 Tab 1]
     * 
     */
    private static final class WhitespaceChars
    {
        static final char NULL = 0x00, NUL = 0x00;
        static final char HORIZONTAL_TAB = 0x09, HT = 0x09;
        static final char LINE_FEED = 0x0A, LF = 0x0A;
        static final char FORM_FEED = 0x0C, FF = 0x0C;
        static final char CARRIAGE_RETURN = 0x0D, CR = 0x0D;
        static final char SPACE = 0x20, SP = 0x20;
    }

    /**
     * 
     * Definition of the list of delimiter characters.
     * [IS0 32000-1:2008: 7.2.2 Tab 2]
     */
    private static final class DelimiterChars
    {
        static final char LEFT_PARENTHESIS = 0x28; // '(';
        static final char RIGHT_PARENTHESIS = 0x29; // ')';
        static final char LESS_THAN_SIGN = 0x3C; // '<';
        static final char GREATER_THAN_SIGN = 0x3E; // '>';
        static final char LEFT_SQUARE_BRACKET = 0x5B; // '[';
        static final char RIGHT_SQUARE_BRACKET = 0x5D; // ']';
        static final char LEFT_CURLY_BRACKET = 0x7B; // '{';
        static final char RIGHT_CURLY_BRACKET = 0x7D; // '}';
        static final char SOLIDUS = 0x2F; // '/';
        static final char PERCENT_SIGN = 0x25; // '%';
    }
    
    /**
     * 
     * Definition for recognition of the keyword endstream.
     * [IS0 32000-1:2008: 7.3.8.1]
     * 
     */    
    private enum EndStream
    {
        out,e1,n,d,s,t,r,e2,a,m;
    }

    /**
     * This is the string constant for endstream.
     */
    private static final String ENDSTREAM_STRING = "endstream";

    /**
     * The character currently read.
     * 
     * As the lexer advances through the PDF the character currently read 
     * might not belong to the current token being processed but be a part
     * of the next token. In order to avoid reading that character twice or 
     * resetting the position the character is held.
     * 
     * This also means that if the lexer is set to a new position 
     * this character has to be reset as this is no longer the next character.
     * 
     */
    private int currentChar;

    /**
     * Signal to read next char instead of taking the current char buffer.
     */
    private boolean needsToRead = true;

    /**
     * The detected tokenType.
     */
    private TokenTypes currentTokenType = TokenTypes.BEGIN_DOCUMENT;

    /**
     * Holds the data read and/or parsed.
     */
    private String currentTokenContent;

    /**
     * The status if the token has been processed. 
     */
    private boolean currentTokenProcessed = true;
    
    /**
     * The starting position of the token within the stream.
     */
    private long tokenOffset = -1;
    
    /**
     * Th end position of the token within the stream.
     */
    private long tokenEnd = -1;


    /**
     * The token buffer will be used to add the token data while parsing the
     * PDF. Moving to the next token the buffer will be cleared and reused.
     */
    private StringBuilder tokenBuffer = new StringBuilder();


    /**
     * The eventType pending next. This might have been found by processing the
     * data for the current event.
     */
    private TokenTypes nextTokenType = TokenTypes.UNKNOWN;

    /**
     * Holds the data read for the next event. This might have been found by
     * processing the data for the current event.
     */
    private String nextTokenData;


    /**
     * The status if the token has been processed. 
     */
    private boolean nextTokenProcessed = true;


    /**
     * Skip whitespace. 
     * 
     * For most parts of parsing PDF whitespace between tokens is treated as being
     * insignificant. The only part where this is not true is the parsing of an Xref Table.
     */
    private boolean skipWhitespace = false;


    /**
     * Skip EOL. 
     * 
     * For most parts of parsing PDF EOL between tokens is treated as being
     * insignificant. The only part where this is not true is the parsing of an Xref Table.
     */
    private boolean skipEOL = false;


    /**
     * The instance of the pdf stream assigned to the lexer.
     */
    private RandomAccessRead stream = null;


    /**
     * Creates a pdf lexer for the pdf stream specified using the stream.
     * 
     * @param sourceStream instance of the stream assigned to the lexer 
     * rather than a regular file, or cannot be opened for any other reason.
     */
    public PDFLexer(RandomAccessRead sourceStream)
    {
        this.stream = sourceStream;
    }
    
    /**
     * Returns the next event.
     * 
     * <p>The events represent most of the basic PDF Objects as specified in
     * ISO-32000-1:2008 7.2. and 7.3. These are
     * </p>
     * 
     * <ul>
     * <li>EOL [IS0 32000-1:2008: 7.2.2 Tab 1]</li>
     * <li>Whitespace [IS0 32000-1:2008: 7.2.2 Tab 1]</li>
     * <li>Comment [IS0 32000-1:2008: 7.2.3]</li>
     * <li>Boolean Objects [IS0 32000-1:2008: 7.3.2]</li>
     * <li>Numeric Objects [IS0 32000-1:2008: 7.3.3]</li>
     * <li>String Objects [IS0 32000-1:2008: 7.3.4]</li>
     * <li>Name Object [IS0 32000-1:2008: 7.3.5]</li>
     * <li>Keyword [IS0 32000-1:2008: 7.3]</li>
     * <li>Array Objects [IS0 32000-1:2008: 7.3.6]</li>
     * <li>Dictionary Objects [IS0 32000-1:2008: 7.3.7]</li>
     * <li>Stream Objects [IS0 32000-1:2008: 7.3.8]</li>
     * <li>Null Objects [IS0 32000-1:2008: 7.3.9]</li>
     * <li>Unknown</li>
     * </ul>
     * 
     * <p>In addition events are delivered for:</>
     * 
     * <ul>
     * <li>start and end of document</li>
     * <li>start and end of a dictionary object</li>
     * <li>start and end of a array object<li>
     * </ul>
     * 
     * <p>Indirect Objects need to be build based on the basic event types.</p>
     * 
     * @return type of the next event.
     * @throws IOException
     *             if an I/O error occurs. Not thrown if end-of-file has been
     *             reached.
     */
    public TokenTypes next() throws IOException
    {
        // ensure the token processing has been finished
        // will be done if getToken() has been called or 
        // within some events
        // otherwise skip over 
        if (!currentTokenProcessed)
        {
            skipTokenProcessing();
        }

        if (nextTokenType != TokenTypes.UNKNOWN) 
        {
            currentTokenType = nextTokenType;
            currentTokenContent = nextTokenData;
            currentTokenProcessed = nextTokenProcessed;

            nextTokenType = TokenTypes.UNKNOWN;
            nextTokenData = null;
            nextTokenProcessed = false;

        }
        else 
        {
            currentTokenType = parseForNextToken();
        }
        return currentTokenType;
    }

    /**
     * Check if there are more events.
     * 
     * @return true if the event reader has more events, false otherwise
     */
    public boolean hasNext()
    {
        return this.currentTokenType != TokenTypes.END_DOCUMENT && this.nextTokenType != TokenTypes.END_DOCUMENT;
    }

    /**
     * Get the type of the current token.
     * 
     * @return the current TokenType.
     */
    public TokenTypes getTokenType()
    {
        return currentTokenType;
    }

    /**
     * Set the type of the current token.
     * 
     * <p>This can be used by parsers extending this class to eventTypes which
     * can be determined at a later parsing stage and are not detected from this
     * class.
     * 
     * @param tokenType the TokenType to set.
     */
    public void setTokenType(TokenTypes tokenType)
    {
        this.currentTokenType = tokenType;
    }

    /**
     * Get the token data.
     * 
     * After an event occurred the cursor is at the start of the token and the
     * reminder of the token has not been processed. An initial call will start
     * reading the data and building the tokens content. Subsequent calls for
     * the same token will take the cached data. When the cursor is forwarded to
     * the next event the data for the previous event is lost. It's up to the
     * calling application to persist the data if needed.
     * 
     * @return the token data.
     * @throws IOException
     *             if an I/O error occurs. Not thrown if end-of-file has been
     *             reached.
     */
    public Token getToken() throws IOException
    {
        if (!currentTokenProcessed)
        {
            finishTokenProcessing();
        }
        
        Token token = new Token(
                currentTokenType, 
                currentTokenContent,
                tokenOffset,
                tokenEnd);
        return token;
    }
    
    /**
     * Get the start postion of the current token.
     * 
     * @return the start position of the current token.
     */
    public long getTokenStart()
    {
        return tokenOffset;
    }

    /**
     * Toggles if whitespace encountered between tokens should be ignored.
     * 
     * @param skipWhitespace true for not returning tokens representing whitespace characters.
     */
    public void setSkipWhitespace(boolean skipWhitespace)
    {
        this.skipWhitespace = skipWhitespace;
    }

    /**
     * State of whitespace handling.
     * 
     * @return true if whitespace is ignored.
     */
    public boolean isSkipWhitespace()
    {
        return skipWhitespace;
    }

    /**
     * Toggles if EOL characters encountered between tokens should be ignored.
     * 
     * @param skipEOL true for not returning tokens representing EOL characters.
     */
    public void setSkipEOL(boolean skipEOL)
    {
        this.skipEOL = skipEOL;
    }

    /**
     * State of whitespace handling.
     * 
     * @return true if EOL is ignored.
     */
    public boolean isSkipEOL()
    {
        return skipEOL;
    }

    // -------------------------------------------------------------------------
    // IO helpers
    // -------------------------------------------------------------------------

    /**
     * Seek to a position in the data.
     *
     * @param position The position to seek to.
     * @throws IOException If there is an error while seeking.
     */
    public void seek(long position) throws IOException 
    {
        // reset the char held in the buffer to 
        // reread after repositioning within the stream
        needsToRead = true;
        stream.seek(position);

        if (position == 0)
        {
            currentTokenType = TokenTypes.BEGIN_DOCUMENT;
        }
        else
        {
            currentTokenType = TokenTypes.UNKNOWN;
        }
        currentTokenContent = null;
        currentTokenProcessed = false;
        nextTokenType = TokenTypes.UNKNOWN;
        nextTokenData = null;
        nextTokenProcessed = false;
        tokenOffset = -1;
        tokenEnd = -1;
    }

    /**
     * Reads a byte of data.
     * 
     * @return the next byte of data, or -1 if the end of the file has been
     *         reached.
     * @throws IOException if an I/O error occurs. Not thrown if end-of-file has
     *         been reached.
     */
    public int read() throws IOException
    {
        return stream.read();
    }
    
    /**
     * Get the current position in the input stream.
     * 
     * @return the current position in the input stream.
     * @throws IOException
     */
    public long getPositon() throws IOException
    {
        return stream.getPosition();
    }


    // -------------------------------------------------------------------------
    // Tests for character classes
    // -------------------------------------------------------------------------

    /**
     * This will tell if the next byte to be read is an end of line byte. [ISO
     * 32000-1:2008: 7.2.2 Tab 1]
     * 
     * @param c
     *            the character to check against end of line.
     * @return true if the next byte is 0x0A or 0x0D.
     */
    private static boolean isEOL(final int c)
    {
        return c == WhitespaceChars.LF || c == WhitespaceChars.CR;
    }

    /**
     * This will tell if the next byte is whitespace or not. [IS0 32000-1:2008:
     * 7.2.2 Tab 1]
     * 
     * @param c
     *            the character to check against whitespace.
     * @return true if the next byte in the stream is a whitespace character.
     */
    protected static boolean isWhitespace(final int c)
    {
        return c == WhitespaceChars.NUL || c == WhitespaceChars.HT
                || c == WhitespaceChars.LF || c == WhitespaceChars.FF
                || c == WhitespaceChars.CR || c == WhitespaceChars.SP;
    }

    /**
     * This will tell if the next byte is a delimiter or not. [IS0 32000-1:2008:
     * 7.2.2 Tab 1]
     * 
     * @param c
     *            the character to check against whitespace.
     * @return true if the next byte in the stream is a delimiter character.
     */
    private static boolean isDelimiter(final int c)
    {
        return c == DelimiterChars.LEFT_PARENTHESIS
                || c == DelimiterChars.RIGHT_PARENTHESIS
                || c == DelimiterChars.LESS_THAN_SIGN
                || c == DelimiterChars.GREATER_THAN_SIGN
                || c == DelimiterChars.LEFT_SQUARE_BRACKET
                || c == DelimiterChars.RIGHT_SQUARE_BRACKET
                || c == DelimiterChars.SOLIDUS
                || c == DelimiterChars.PERCENT_SIGN;
    }

    /**
     * This will tell if the next byte is a permitted numeric character.
     * 
     * @param c
     *            the character to check against whitespace.
     * @return true if the next byte in the stream is a numeric character.
     */
    private static boolean isNumeric(final int c)
    {
        return c == '0' || c == '1' || c == '2' || c == '3' || c == '4'
                || c == '5' || c == '6' || c == '7' || c == '8' || c == '9'
                || c == '.' || c == '+' || c == '-';
    }

    // -------------------------------------------------------------------------
    // Handling of the different stage transitions
    // -------------------------------------------------------------------------

    private TokenTypes parseForNextToken() throws IOException
    {

        if (!currentTokenProcessed)
        {
            skipTokenProcessing();
        }

        if (needsToRead)
        {
            currentChar = read();
            needsToRead = false;
        }

        tokenOffset = stream.getPosition() -1;
        
        switch (currentChar)
        {

        // EOL characters
        case WhitespaceChars.CARRIAGE_RETURN:
        case WhitespaceChars.LINE_FEED:
            currentTokenProcessed = false;
            currentTokenType = TokenTypes.EOL;

            if (isSkipEOL())
            {
                currentTokenType = parseForNextToken();
            }

            break;

            // Whitespace
        case WhitespaceChars.NULL:
        case WhitespaceChars.HORIZONTAL_TAB:
        case WhitespaceChars.FORM_FEED:
        case WhitespaceChars.SPACE:
            currentTokenProcessed = false;
            currentTokenType = TokenTypes.WHITESPACE;

            if (isSkipWhitespace())
            {
                currentTokenType = parseForNextToken();
            }

            break;

            // Comment
        case DelimiterChars.PERCENT_SIGN:
            currentTokenProcessed = false;
            currentTokenType = TokenTypes.COMMENT;
            break;

            // Numeric Objects
        case '0': case '1': case '2': case '3':
        case '4': case '5': case '6': case '7':
        case '8': case '9': 
        case '+': case '-': case '.':
            currentTokenProcessed = false;
            currentTokenType = TokenTypes.NUMBER;
            break;

            // Literal String Objects
        case DelimiterChars.LEFT_PARENTHESIS:
            currentTokenProcessed = false;
            currentTokenType = TokenTypes.LITERAL;
            break;

            // Name Object
        case DelimiterChars.SOLIDUS:
            currentTokenProcessed = false;
            currentTokenType = TokenTypes.NAME;
            break;

            // Array start
        case DelimiterChars.LEFT_SQUARE_BRACKET:
            currentChar = read();
            currentTokenProcessed = true;
            currentTokenContent = null;
            currentTokenType = TokenTypes.ARRAY_START;
            break;

            // Array end
        case DelimiterChars.RIGHT_SQUARE_BRACKET:
            currentChar = read();
            currentTokenProcessed = true;
            currentTokenContent = null;
            currentTokenType = TokenTypes.ARRAY_END;
            break;

            // Dictionary Object or Hex String Object
        case DelimiterChars.LESS_THAN_SIGN:
            // This could be either the start of a
            // Dictionary [IS0 32000-1:2008: 7.3.7] or a
            // Hexadecimal String [IS0 32000-1:2008: 7.3.4]
            // so we need to read the next byte to make
            // a decision
            int nextChar = read();

            if (nextChar == DelimiterChars.LESS_THAN_SIGN)
            {
                // Start of a Dictionary Object
                currentTokenProcessed = true;
                currentTokenContent = tokenBuffer.toString();
                currentTokenType = TokenTypes.DICTIONARY_START;
                tokenEnd = stream.getPosition()-1;
                needsToRead = true;
            }
            else
            {
                // Start of a Hex String Object
                currentTokenProcessed = false;
                currentChar = nextChar;
                currentTokenType = TokenTypes.HEX;
            }
            break;

            // Dictionary Object end
        case DelimiterChars.GREATER_THAN_SIGN:
            // we need to check for the second
            // ClosingAngleBracket for a valid
            // dictionary end
            clearTokenBuffer();
            nextChar = read();

            if (nextChar == DelimiterChars.GREATER_THAN_SIGN)
            {
                currentTokenProcessed = true;
                currentTokenType = TokenTypes.DICTIONARY_END;
                currentTokenContent = null;
                tokenEnd = stream.getPosition()-1;
                needsToRead = true;
            }
            else 
            {
                // TODO: that shouldn't happen in a valid PDF
                // enhance to support non valid PDFs
            }

            break;

        case -1:
            currentTokenProcessed = true;
            currentTokenContent = null;
            currentTokenType = TokenTypes.END_DOCUMENT;
            break;

        default: // Keyword
            currentTokenType = TokenTypes.KEYWORD;
            currentTokenProcessed = false;

            // For keywords the content has to be parsed as this
            // could be a BOOLEAN, NULL, STREAM_OBJECT_START,
            // or something else.
            // As BOOLEAN, NULL and STREAMs are base objects the decision
            // has to be made

            String tokenContent = getToken().getContent();
            
            if ("stream".equals(tokenContent))
            {
                // This is the start of the stream object. The other
                // related eventTypes STREAM_OBJECT and STREAM_OBJECT_END
                // will be handled by transition from START to END
                //
                // The different parts of a stream will be treated
                // separately as to check for conformance with the PDF spec
                // the keywords need to be inspected 
                currentTokenType = TokenTypes.STREAM_START;

                // for the begin of a stream continue to read into 
                // the next bytes as these are part of the keyword.
                finishTokenProcessing();
            }
            else if ("true".equals(tokenContent))
            {
                currentTokenType = TokenTypes.BOOLEAN;
            }
            else if ("false".equals(tokenContent)) 
            {
                currentTokenType = TokenTypes.BOOLEAN;
            }
            else if ("null".equals(tokenContent))
            {
                currentTokenType = TokenTypes.NULL;
            }

            break;
        }

        return currentTokenType;
    }

    /**
     * Clear the tokenBuffer.
     * 
     */
    private void clearTokenBuffer()
    {
        tokenBuffer.setLength(0);

    }

    /**
     * Finish the processing of the data for the token started by the event.
     * 
     * As the processing of the data is deferred until the data is requested
     * make sure that we parse the data until we get to the next event.
     * 
     * This method is called by {@link #getToken()}.
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void finishTokenProcessing() throws IOException
    {

        currentTokenProcessed = true;

        switch (currentTokenType)
        {
        case EOL:
            processEOL();
            break;
        case WHITESPACE:
            processWhitespace();
            break;
        case COMMENT:
            processComment();
            break;
        case NUMBER:
            processNumber();
            break;
        case LITERAL:
            processLiteralString();
            break;
        case HEX:
            processHexString();
            break;
        case NAME:
            processName();
            break;
        case KEYWORD:
        case NULL:
        case BOOLEAN:
            processKeyword();
            break;
        case STREAM_START:
            processStreamObjectStart();
            break;
        case STREAM:
            processStream();
            break;
        case STREAM_END:
            break;
        case BEGIN_DOCUMENT:
            currentTokenContent = null;
            break;
        case END_DOCUMENT:
            currentTokenContent = null;
            break;
        case UNKNOWN:
            currentTokenContent = null;
            break;
        default:
            // this will mean that we have an eventType encountered which we do not
            // handle
            throw new IOException("unhandled eventType " + currentTokenType + " encountered at ");
        }
    }

    /**
     * Skip the processing of the data for the token started by the event.
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipTokenProcessing() throws IOException 
    {

        currentTokenProcessed = true;
        currentTokenContent = null;

        switch (currentTokenType)
        {
        case EOL:
            skipEOL();
            break;
        case WHITESPACE:
            skipWhitespace();
            break;
        case COMMENT:
            skipComment();
            break;
        case NUMBER:
            skipNumber();
            break;
        case LITERAL:
            skipLiteralString();
            break;
        case HEX:
            skipHexString();
            break;
        case NAME:
            skipName();
            break;
        case KEYWORD:
        case NULL:
        case BOOLEAN:
            skipKeyword();
            break;
        case STREAM_START:
            skipStreamObjectStart();
            break;
        case STREAM:
            skipStream();
            break;
        case STREAM_END:
            break;
        case BEGIN_DOCUMENT:
            break;
        case END_DOCUMENT:
            break;
        case UNKNOWN:
            break;
        default:
            // this will mean that we have an eventType encountered which we do not
            // handle
            throw new IOException("unhandled eventType " + currentTokenType + " encountered at ");
        }
    }

    // -------------------------------------------------------------------------
    // processing of event types
    // -------------------------------------------------------------------------

    /**
     * Read the EOL data.
     * 
     * EOL consists of the chars LINE FEED and FORM FEED as defined in 
     * [IS0 32000-1:2008: 7.2.2 Tab 1].
     * 
     * For most applications Whitespace and EOL can be treated the same but we
     * make a difference here as to check for conformance to the PDF
     * specification we need to be able to identify both.
     * 
     * @param keepData if true the data will be put into the rawData. if false
     *              data will be skipped. 
     * @return true if the token could be read.
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void processEOL() throws IOException
    {
        clearTokenBuffer();

        while (true)
        {
            tokenBuffer.append((char) currentChar);
            currentChar = read();
            if (currentChar == -1)
            {
                tokenEnd = stream.getPosition() -1;
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (!isEOL(currentChar)) 
            {
                tokenEnd = stream.getPosition() -1;
                break;
            }
        }
        currentTokenContent = tokenBuffer.toString();
    }

    /**
     * Read the whitespace data.
     * 
     * Whitespace consists of all chars as defined in [IS0 32000-1:2008: 7.2.2
     * Tab 1] other than the eol chars LINE FEED and FORM FEED.
     * 
     * For most applications Whitespace and EOL can be treated the same but we
     * make a difference here as to check for conformance to the PDF
     * specification we need to be able to identify both.
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void processWhitespace() throws IOException
    {
        clearTokenBuffer();

        while (true)
        {
            tokenBuffer.append((char) currentChar);
            currentChar = read();
            if (currentChar == -1)
            {
                tokenEnd = stream.getPosition() -1;
                currentTokenContent = tokenBuffer.toString();
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (isEOL(currentChar) || !isWhitespace(currentChar))
            {
                tokenEnd = stream.getPosition() -1;
                break;
            }
        }
        currentTokenContent = tokenBuffer.toString();
    }

    /**
     * Read the comment data.
     * 
     * The comment consists of all characters after the delimiter up to the end
     * of the line. [IS0 32000-1:2008: 7.2.3]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void processComment() throws IOException 
    {
        clearTokenBuffer();

        while (true)
        {
            tokenBuffer.append((char) currentChar);
            currentChar = read();
            if (currentChar == -1)
            {
                tokenEnd = stream.getPosition() -1;
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (isEOL(currentChar))
            {
                tokenEnd = stream.getPosition() -2;
                break;
            }
        }
        currentTokenContent = tokenBuffer.toString();
    }

    /**
     * Read the number data.
     * 
     * [IS0 32000-1:2008: 7.3.3]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void processNumber() throws IOException
    {
        clearTokenBuffer();

        while (true)
        {
            tokenBuffer.append((char) currentChar);
            currentChar = read();
            if (currentChar == -1)
            {
                tokenEnd = stream.getPosition() -1;
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (!isNumeric(currentChar))
            {
                tokenEnd = stream.getPosition() -2;
                break;
            }
        }
        currentTokenContent = tokenBuffer.toString();
    }

    /**
     * Read the literal string data.
     * 
     * The literal string consists of all characters enclosed in parentheses.
     * Any character may appear in a string except unbalanced parentheses. These
     * can be escaped with a backslash (REVERSE SOLIDUS). [IS0 32000-1:2008:
     * 7.2.4.2]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void processLiteralString() throws IOException 
    {
        int level = 1;
        clearTokenBuffer();
        
        while (true)
        {
            currentChar = read();
            
            if (currentChar == -1)
            {
                tokenEnd = stream.getPosition() -1;
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (currentChar == DelimiterChars.LEFT_PARENTHESIS)
            {
                level++;
            }
            else if (currentChar == DelimiterChars.RIGHT_PARENTHESIS)
            {
                level--;
            }
            else if (currentChar == '\\')
            {
                // tokenBuffer.append((char) currentChar);
                currentChar = read();
                              
                switch (currentChar) 
                {
                    case -1:
                        tokenEnd = stream.getPosition() -1;
                        nextTokenType = TokenTypes.END_DOCUMENT;
                        break;
                    case 'n':
                        tokenBuffer.append('\n');
                        break;
                    case 'r':
                        tokenBuffer.append('\r');
                        break;
                    case 't':
                        tokenBuffer.append('\t');
                        break;
                    case 'b':
                        tokenBuffer.append('\b');
                        break;
                    case 'f':
                        tokenBuffer.append('\f');
                        break;
                    case '\\': case '(': case ')':
                        // tokenBuffer.append((char) currentChar);
                        break;
                    // escaped octal value
                    case '0': case '1': case '2': case '3':
                    case '4': case '5': case '6': case '7':
                        int octal = currentChar - '0';
                        
                        currentChar = read();
                        
                        if (currentChar < '0' || currentChar > '7')
                        {
                            tokenBuffer.append((char) octal);
                        }
                        
                        octal = (octal << 3) + currentChar - '0';
                        
                        currentChar = read();
                        
                        if (currentChar < '0' || currentChar > '7')
                        {
                            tokenBuffer.append((char) octal);
                        }
                        
                        octal = (octal << 3) + currentChar - '0';
                        
                        tokenBuffer.append(octal & 0xff);
                    default:
                        if (currentChar == -1) 
                        {
                            tokenEnd = stream.getPosition() -1;
                            nextTokenType = TokenTypes.END_DOCUMENT;
                            level = 0;
                            break;
                        }   
                }
            }

            if (level == 0) 
            {
                tokenEnd = stream.getPosition() -1;
                needsToRead = true; 
                break;
            } 
            else 
            {
                tokenBuffer.append((char) currentChar);
            }
        }
        currentTokenContent = tokenBuffer.toString();
    }

    /**
     * Read the hex string data.
     * 
     * A hexadecimal string shall be written as a sequence of hexadecimal digits
     * (0-9 and either A-F or a-f) encoded as ASCII characters and enclosed
     * within angle brackets (using LESS-THAN SIGN (3Ch) and GREATER- THAN SIGN
     * (3Eh)). [IS0 32000-1:2008: 7.2.4.3]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void processHexString() throws IOException
    {
        clearTokenBuffer();        
        while (true)
        {
            tokenBuffer.append((char) currentChar);
            currentChar = read();
            if (currentChar == -1)
            {
                tokenEnd = stream.getPosition() -1;
                currentTokenContent = tokenBuffer.toString();
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if(currentChar == DelimiterChars.GREATER_THAN_SIGN)
            {
                tokenEnd = stream.getPosition()-1;
                needsToRead = true;
                break;
            }
        }
        currentTokenContent = tokenBuffer.toString();
    }

    /**
     * Read the name objects data.
     * 
     * [IS0 32000-1:2008: 7.3.5]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void processName() throws IOException
    {
        clearTokenBuffer();

        while (true)
        {
            currentChar = read();
            if (currentChar == -1)
            {
                tokenEnd = stream.getPosition() -1;
                currentTokenContent = tokenBuffer.toString();
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if(isDelimiter(currentChar) || isWhitespace(currentChar))
            {
                tokenEnd = stream.getPosition() -2;
                nextTokenType = TokenTypes.UNKNOWN;
                break;
            }
            tokenBuffer.append((char) currentChar);
        }
        currentTokenContent = tokenBuffer.toString();
    }

    /**
     * Read the keyword data.
     * 
     * [IS0 32000-1:2008: 7.3]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void processKeyword() throws IOException
    {
        clearTokenBuffer();

        while (true)
        {
            tokenBuffer.append((char) currentChar);
            currentChar = read();
            if (currentChar == -1)
            {
                tokenEnd = stream.getPosition() -1;
                currentTokenContent = tokenBuffer.toString();
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (isDelimiter(currentChar) || isWhitespace(currentChar))
            {
                tokenEnd = stream.getPosition() -2;
                break;
            }
        }
        currentTokenContent = tokenBuffer.toString();
    }


    /**
     * Read the stream object start data.
     * 
     * The keyword 'stream' that follows the stream dictionary shall be
     * followed by an end-of-line marker consisting of either a
     * CARRIAGE RETURN and a LINE FEED or just a LINE FEED, and not by a
     * CARRIAGE RETURN alone.
     * [IS0 32000-1:2008: 7.3.8.1]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void processStreamObjectStart() throws IOException
    {
        clearTokenBuffer();
        // the keyword 'stream' has been read already - append that
        tokenBuffer.append(currentTokenContent);

        // check if the end-of-line-marker is CR/LF
        if (currentChar == WhitespaceChars.CARRIAGE_RETURN)
        {
            //append the data
            tokenBuffer.append((char) currentChar);                
            currentChar = read();
            if (currentChar == WhitespaceChars.LINE_FEED)
            {
                //append the data
                tokenBuffer.append((char) currentChar);                
                nextTokenType = TokenTypes.STREAM;
                nextTokenProcessed = false;
                tokenEnd = stream.getPosition()-1;
                needsToRead = true;
            }
            // check if the end-of-line-marker is LF
        }
        else if (currentChar == WhitespaceChars.LINE_FEED)
        {
            //append the data and finish
            tokenBuffer.append((char) currentChar);                
            nextTokenType = TokenTypes.STREAM;
            nextTokenProcessed = false;
            tokenEnd = stream.getPosition()-1;
            needsToRead = true;
        }
        else if (currentChar == -1)
        {
            // this shouldn't happen but simply pass the tokens and
            // let the parser deal with the issue
            tokenEnd = stream.getPosition() -1;
            nextTokenType = TokenTypes.END_DOCUMENT;
            needsToRead = false;
        }
        currentTokenContent = tokenBuffer.toString();
    }


    /**
     * Read the stream object data.
     * 
     * [IS0 32000-1:2008: 7.3.8]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void processStream() throws IOException
    {

        clearTokenBuffer();
        EndStream endStream = EndStream.out;

        while (true)
        {
            currentChar = read();

            if (currentChar == -1)
            {
                tokenEnd = stream.getPosition() -1;
                currentTokenContent = tokenBuffer.toString();
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }

            tokenBuffer.append((char) currentChar);

            // test for endstream
            if ('e' == currentChar && endStream == EndStream.out)
            {
                endStream = EndStream.e1;
            }
            else if ('n' == currentChar && endStream == EndStream.e1)
            {
                endStream = EndStream.n;
            }
            else if ('d' == currentChar && endStream == EndStream.n)
            {
                endStream = EndStream.d;
            }
            else if ('s' == currentChar && endStream == EndStream.d)
            {
                endStream = EndStream.s;
            }
            else if ('t' == currentChar && endStream == EndStream.s)
            {
                endStream = EndStream.t;
            }
            else if ('r' == currentChar && endStream == EndStream.t)
            {
                endStream = EndStream.r;
            }
            else if ('e' == currentChar && endStream == EndStream.r)
            {
                endStream = EndStream.e2;
            }
            else if ('a' == currentChar && endStream == EndStream.e2)
            {
                endStream = EndStream.a;
            }
            else if ('m' == currentChar && endStream == EndStream.a)
            {
                endStream = EndStream.m;

                // Strip off endstream from the data. If there is a EOL
                // character at the end of the stream data just before
                // endstream this will not be part of the stream data.
                // 
                // [IS0 32000-1:2008: 7.3.8.1]: There should be an
                // end-of-line marker after the data and before endstream;
                // this marker shall not be included in the stream length.
                //
                if (isEOL(tokenBuffer.charAt(tokenBuffer.length() - 10)))
                {
                    currentTokenContent = tokenBuffer.substring(0,tokenBuffer.length()-10);
                }
                else
                {
                    currentTokenContent = tokenBuffer.substring(0,tokenBuffer.length()-9);
                }

                nextTokenType = TokenTypes.STREAM_END;
                nextTokenData = ENDSTREAM_STRING;
                nextTokenProcessed = true;
                tokenEnd = stream.getPosition()-1;
                needsToRead = true;
                break;
            }
            else
            {
                endStream = EndStream.out;
            }
        }
    }


    // -------------------------------------------------------------------------
    // skipping of event types
    // -------------------------------------------------------------------------

    /**
     * Read the EOL data.
     * 
     * EOL consists of the chars LINE FEED and FORM FEED as defined in [IS0
     * 32000-1:2008: 7.2.2 Tab 1].
     * 
     * For most applications Whitespace and EOL can be treated the same but we
     * make a difference here as to check for conformance to the PDF
     * specification we need to be able to identify both.
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipEOL() throws IOException
    {
        while (true)
        {
            currentChar = read();
            if (currentChar == -1)
            {
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (!isEOL(currentChar))
            {
                break;
            }
        }
    }

    /**
     * Read the whitespace data.
     * 
     * Whitespace consists of all chars as defined in [IS0 32000-1:2008: 7.2.2
     * Tab 1] other than the eol chars LINE FEED and FORM FEED.
     * 
     * For most applications Whitespace and EOL can be treated the same but we
     * make a difference here as to check for conformance to the PDF
     * specification we need to be able to identify both.
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipWhitespace() throws IOException
    {
        while (true)
        {
            currentChar = read();
            if (currentChar == -1)
            {
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (isEOL(currentChar) || !isWhitespace(currentChar))
            {
                break;
            }
        }
    }

    /**
     * Skip the comment data.
     * 
     * The comment consists of all characters after the delimiter up to the end
     * of the line. [IS0 32000-1:2008: 7.2.3]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipComment() throws IOException
    {
        while (true)
        {
            currentChar = read();
            if (currentChar == -1)
            {
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (isEOL(currentChar))
            {
                break;
            }
        }
    }

    /**
     * Skip the number data.
     * 
     * [IS0 32000-1:2008: 7.3.3]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipNumber() throws IOException
    {
        while (true)
        {
            currentChar = read();
            if (currentChar == -1)
            {
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (!isNumeric(currentChar))
            {
                break;
            }
        }
    }

    /**
     * Skip the literal string data.
     * 
     * The literal string consists of all characters enclosed in parentheses.
     * Any character may appear in a string except unbalanced parentheses. These
     * can be escaped with a backslash (REVERSE SOLIDUS). [IS0 32000-1:2008:
     * 7.2.4.2]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipLiteralString() throws IOException
    {
        int level = 1;

        while (true)
        {
            currentChar = read();
            if (currentChar == -1)
            {
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (currentChar == DelimiterChars.LEFT_PARENTHESIS)
            {
                level++;
            }
            else if (currentChar == DelimiterChars.RIGHT_PARENTHESIS)
            {
                level--;
            }
            else if (currentChar == '\\')
            {
                currentChar = read();
            }

            if (level == 0)
            {
                needsToRead = true;
                break;
            }
        }
    }

    /**
     * Skip the hex string data.
     * 
     * A hexadecimal string shall be written as a sequence of hexadecimal digits
     * (0-9 and either A-F or a-f) encoded as ASCII characters and enclosed
     * within angle brackets (using LESS-THAN SIGN (3Ch) and GREATER- THAN SIGN
     * (3Eh)). [IS0 32000-1:2008: 7.2.4.3]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipHexString() throws IOException
    {
        while (true)
        {
            currentChar = read();
            if (currentChar == -1)
            {
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (currentChar == DelimiterChars.GREATER_THAN_SIGN)
            {
                needsToRead = true;
                break;
            }
        }
    }

    /**
     * Skip the name objects data.
     * 
     * [IS0 32000-1:2008: 7.3.5]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipName() throws IOException
    {
        while (true)
        {
            currentChar = read();
            if (currentChar == -1)
            {
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (isDelimiter(currentChar) || isWhitespace(currentChar))
            {
                break;
            }
        }
    }

    /**
     * Skip the keyword data.
     * 
     * [IS0 32000-1:2008: 7.3]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipKeyword() throws IOException
    {
        while (true)
        {
            currentChar = read();
            if (currentChar == -1)
            {
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }
            else if (isDelimiter(currentChar) || isWhitespace(currentChar))
            {
                break;
            }
        }
    }


    /**
     * Skip the stream object start data.
     * 
     * The keyword 'stream' that follows the stream dictionary shall be
     * followed by an end-of-line marker consisting of either a
     * CARRIAGE RETURN and a LINE FEED or just a LINE FEED, and not by a
     * CARRIAGE RETURN alone.
     * [IS0 32000-1:2008: 7.3.8.1]
     * 
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipStreamObjectStart() throws IOException
    {
        // check if the end-of-line-marker is CR/LF
        if (currentChar == WhitespaceChars.CARRIAGE_RETURN)
        {
            //append the data
            currentChar = read();
            if (currentChar == WhitespaceChars.LINE_FEED)
            {
                //append the data
                nextTokenType = TokenTypes.STREAM;
                nextTokenProcessed = false;
                needsToRead = true;
            }
            // check if the end-of-line-marker is LF
        }
        else if (currentChar == WhitespaceChars.LINE_FEED)
        {
            //append the data and finish
            nextTokenType = TokenTypes.STREAM;
            nextTokenProcessed = false;
            needsToRead = true;
        }
        else if (currentChar == -1)
        {
            // this shouldn't happen but simply pass the tokens and
            // let the parser deal with the issue
            nextTokenType = TokenTypes.END_DOCUMENT;
            needsToRead = false;
        }
    }

    /**
     * Skip the stream object data.
     * 
     * [IS0 32000-1:2008: 7.3.8]
     * 
     * @param keepData if true the data will be put into the rawData. if false
     *              data will be skipped. 
     * @return true if the token could be read.
     * @throws IOException
     *             in case there is an issue reading the underlying data
     */
    private void skipStream() throws IOException
    {
        EndStream endStream = EndStream.out;

        while (true)
        {
            currentChar = read();

            if (currentChar == -1)
            {
                currentTokenContent = tokenBuffer.toString();
                nextTokenType = TokenTypes.END_DOCUMENT;
                break;
            }

            // test for endstream
            if ('e' == currentChar && endStream == EndStream.out)
            {
                endStream = EndStream.e1;
            }
            else if ('n' == currentChar && endStream == EndStream.e1)
            {
                endStream = EndStream.n;
            }
            else if ('d' == currentChar && endStream == EndStream.n)
            {
                endStream = EndStream.d;
            }
            else if ('s' == currentChar && endStream == EndStream.d)
            {
                endStream = EndStream.s;
            }
            else if ('t' == currentChar && endStream == EndStream.s)
            {
                endStream = EndStream.t;
            }
            else if ('r' == currentChar && endStream == EndStream.t)
            {
                endStream = EndStream.r;
            }
            else if ('e' == currentChar && endStream == EndStream.r)
            {
                endStream = EndStream.e2;
            }
            else if ('a' == currentChar && endStream == EndStream.e2)
            {
                endStream = EndStream.a;
            }
            else if ('m' == currentChar && endStream == EndStream.a)
            {
                endStream = EndStream.m;

                nextTokenType = TokenTypes.STREAM_END;
                nextTokenData = ENDSTREAM_STRING;
                nextTokenProcessed = true;
                needsToRead = true;
                break;
            }
            else
            {
                endStream = EndStream.out;
            }
        }
    }
}
