/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.pdfbox.pdfparser.tokens;

/**
 * Representation of the different types of tokens as defined in the PDF
 * spec. [IS0 32000-1:2008: 7.2.3, 7.3]
 * 
 * @author Maruan Sahyoun (sahyoun at fileaffairs.de)
 * @version $Revision: 1.0 $
 */

public enum TokenTypes 
{
    /**
     * Indicates an event is an eol character/characters.
     * [IS0 32000-1:2008: 7.2.2 Tab 1]
     */
    EOL,

    /**
     * Indicates an event is a whitespace character/characters.
     * [IS0 32000-1:2008: 7.2.2 Tab 1]
     */
    WHITESPACE,
    
    /**
     * Indicates an event is a comment.
     * [IS0 32000-1:2008: 7.2.3]
     */
    COMMENT,
    
    /**
     * Indicates an event is a boolean.
     * [IS0 32000-1:2008: 7.3.2]
     */    
    BOOLEAN,

    /**
     * Indicates an event is a number.
     * [IS0 32000-1:2008: 7.3.3]
     */  
    NUMBER,
    
    /**
     * Indicates an event is a number.
     * [IS0 32000-1:2008: 7.3.3]
     */  
    INTEGER,

    /**
     * Indicates an event is a number.
     * [IS0 32000-1:2008: 7.3.3]
     */  
    REAL,

    /**
     * Indicates an event is a literal string.
     * [IS0 32000-1:2008: 7.3.4.2]
     */  
    LITERAL,
    
    /**
     * Indicates an event is a hex string.
     * [IS0 32000-1:2008: 7.3.4.3]
     */      
    HEX,
    
    /**
     * Indicates an event is a name object.
     * [IS0 32000-1:2008: 7.3.5]
     */  
    NAME,
    
    /**
     * Indicates an event is an array start.
     * [IS0 32000-1:2008: 7.3.6]
     */
    ARRAY_START,
    
    /**
     * Indicates an event is an array end.
     * [IS0 32000-1:2008: 7.3.6]
     */
    ARRAY_END,
    
    /**
     * Indicates an event is an array object.
     * [IS0 32000-1:2008: 7.3.6]
     */
    ARRAY,    

    /**
     * Indicates an event is a dictionary start.
     * [IS0 32000-1:2008: 7.3.7]
     */
    DICTIONARY_START,
    
    /**
     * Indicates an event is a dictionary end.
     * [IS0 32000-1:2008: 7.3.7]
     */
    DICTIONARY_END,
    
    /**
     * Indicates an event is a dictionary object.
     * [IS0 32000-1:2008: 7.3.7]
     */
    DICTIONARY,
    
    /**
     * Indicates an event is a stream start.
     * [IS0 32000-1:2008: 7.3.8]
     */
    STREAM_START,
    
    /**
     * Indicates an event is a stream end.
     * [IS0 32000-1:2008: 7.3.8]
     */
    STREAM_END,

    /**
     * Indicates an event is a stream object.
     * [IS0 32000-1:2008: 7.3.8]
     */
    STREAM,

    /**
     * Indicates an event is a null object.
     * [IS0 32000-1:2008: 7.3.9]
     */
    NULL,

    /**
     * Indicates an event is an indirect object start.
     * [IS0 32000-1:2008: 7.3.10]
     */
    INDIRECT_START,
    
    /**
     * Indicates an event is an indirect object end.
     * [IS0 32000-1:2008: 7.3.10]
     */
    INDIRECT_END,
    
    /**
     * Indicates an event is an indirect object.
     * [IS0 32000-1:2008: 7.3.10]
     */
    INDIRECT,
    
    /**
     * Indicates an event is a keyword.
     * [IS0 32000-1:2008: 7.3]
     */
    KEYWORD,

    /**
     * Indicates an event is a begin of document.
     * 
     * This event indicates the start of the processing. It doesn't necessarily
     * indicate the begin of the complete document as only sub sections can be
     * processed.
     * 
     * [IS0 32000-1:2008: 7.2.2 Tab 1]
     */
    BEGIN_DOCUMENT,
    
    /**
     * Indicates an event is an end of document.
     * 
     * This event indicates the end of the processing by reaching the end of the
     * readable data. It doesn't necessarily indicate the end of the complete
     * document as only sub sections can be processed.
     * 
     * [IS0 32000-1:2008: 7.2.2 Tab 1]
     */
    END_DOCUMENT,

    /**
     * Indicates an event is of an unknown type.
     */
    UNKNOWN
}