/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.pdfbox.pdfparser.tokens;

import org.junit.Assert;
import org.junit.Test;

import org.apache.pdfbox.io.RandomAccessByteArray;
import org.apache.pdfbox.io.RandomAccessRead;

/**
 * Test suite for the PDFLexer.
 * 
 * <p>The events represent most of the basic PDF Objects as specified in
 * ISO-32000-1:2008 7.2. and 7.3. These are
 * </p>
 * 
 * <ul>
 * <li>EOL [IS0 32000-1:2008: 7.2.2 Tab 1]</li>
 * <li>Whitespace [IS0 32000-1:2008: 7.2.2 Tab 1]</li>
 * <li>Comment [IS0 32000-1:2008: 7.2.3]</li>
 * <li>Boolean Objects [IS0 32000-1:2008: 7.3.2]</li>
 * <li>Numeric Objects [IS0 32000-1:2008: 7.3.3]</li>
 * <li>String Objects [IS0 32000-1:2008: 7.3.4]</li>
 * <li>Name Object [IS0 32000-1:2008: 7.3.5]</li>
 * <li>Keyword [IS0 32000-1:2008: 7.3]</li>
 * <li>Array Objects [IS0 32000-1:2008: 7.3.6]</li>
 * <li>Dictionary Objects [IS0 32000-1:2008: 7.3.7]</li>
 * <li>Stream Objects [IS0 32000-1:2008: 7.3.8]</li>
 * <li>Null Objects [IS0 32000-1:2008: 7.3.9]</li>
 * <li>Unknown</li>
 * </ul>
 * 
 * <p>In addition events are delivered for:</>
 * 
 * <ul>
 * <li>start and end of document</li>
 * <li>start and end of a dictionary object</li>
 * <li>start and end of a array object<li>
 * </ul>
 * 
 * @version $Revision: 1.0 $
 */
public class TestPDFLexer
{
    // helper function to check for token types
    private static void checkTokenTypes(String data, Boolean skipWhitespace, TokenTypes... validTokens) throws Exception
    {
        // there will be two runs - one consuming the token and the other one skipping through them
        // they both need to get the same tokens

        // First pass for skipping the token data
        RandomAccessRead source = new RandomAccessByteArray(data.getBytes());
        PDFLexer lexer = new PDFLexer(source);
        if (skipWhitespace)
        {
            lexer.setSkipWhitespace(true);
        }

        for (int i = 0; i < validTokens.length; i++)
        {
            lexer.next();
            Assert.assertEquals("Token " + i, validTokens[i], lexer.getTokenType());
        }

        source.close();

        // Second pass for reading the token data
        source = new RandomAccessByteArray(data.getBytes());
        lexer = new PDFLexer(source);
        if (skipWhitespace)
        {
            lexer.setSkipWhitespace(true);
        }

        for (int j = 0; j < validTokens.length; j++)
        {
            lexer.next();
            lexer.getToken();
            Assert.assertEquals("Token " + j, validTokens[j], lexer.getTokenType());
        }

        source.close();
    }

    // helper function to check for token content comparing the representation
    // within the pdf (pdfString) to the parsed result as it's represented 
    // internally (parsedString)
    private static void checkTokenContent(String pdfString, String parsedString) throws Exception
    {
        // First pass for skipping the token data
        RandomAccessRead source = new RandomAccessByteArray(pdfString.getBytes());
        PDFLexer lexer = new PDFLexer(source);
        // get beyond the START_DOCUMENT token
        lexer.next();
        // check the content
        Assert.assertEquals(parsedString, lexer.getToken().getContent());
        // ensure that the next token is EOD
        Assert.assertEquals(TokenTypes.END_DOCUMENT, lexer.next());
        source.close();
    }
    
    
    /*
     * Tests for basic token types
     */

    /**
     * Test EOL parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeEOL() throws Exception
    {
        checkTokenTypes(
                "\n",
                false,
                TokenTypes.EOL,
                TokenTypes.END_DOCUMENT
                );
        checkTokenTypes(
                "\r",
                false,
                TokenTypes.EOL,
                TokenTypes.END_DOCUMENT
                );
        checkTokenTypes(
                "\r\n",
                false,
                TokenTypes.EOL,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test whitespace parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeWhitespace() throws Exception
    {
        checkTokenTypes(
                " ",
                false,
                TokenTypes.WHITESPACE,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test comment parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeComment() throws Exception
    {
        checkTokenTypes(
                "%PDF-1.4",
                false,
                TokenTypes.COMMENT,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test keyword parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeKeyword() throws Exception
    {
        checkTokenTypes(
                "obj",
                false,
                TokenTypes.KEYWORD,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test boolean parsing.
     * 
     * Ensure that true and false are treated as a BOOLEAN event type
     * and not handled as KEYWORD.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeBoolean() throws Exception
    {

        checkTokenTypes(
                "true",
                false,
                TokenTypes.BOOLEAN,
                TokenTypes.END_DOCUMENT
                );

        checkTokenTypes(
                "false",
                false,
                TokenTypes.BOOLEAN,
                TokenTypes.END_DOCUMENT
                );
        
        checkTokenTypes(
                "true false",
                false,
                TokenTypes.BOOLEAN,
                TokenTypes.WHITESPACE,
                TokenTypes.BOOLEAN,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test number parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeNumber() throws Exception
    {
        checkTokenTypes(
                "3871",
                false,
                TokenTypes.NUMBER,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test literal string parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeLiteral() throws Exception
    {
        checkTokenTypes(
                "(Test)",
                false,
                TokenTypes.LITERAL,
                TokenTypes.END_DOCUMENT
                );
        // check for nested parts
        checkTokenTypes(
                "(Adobe InDesign CS5 \\(7.0.4\\))",
                false,
                TokenTypes.LITERAL,
                TokenTypes.END_DOCUMENT
                );

    }
    
    /**
     * Test repositioning in stream.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void literalStringParsing() throws Exception
    {
        checkTokenContent("(Test)", "Test");
        // escaped parentheses
        checkTokenContent("(Adobe InDesign CS5 \\(7.0.4\\))", "Adobe InDesign CS5 (7.0.4)");
        // octal with additional number at the end which should be treated as individual
        // character for the string
        checkTokenContent("(Test \120 \1210)", "Test P Q0");
        // octals without separator in between
        checkTokenContent("(\124\145\163\164)", "Test");
        // escaped special character
        checkTokenContent("(Test \n Test)", "Test \n Test");
    }
    
    /**
     * Test hex string parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeHex() throws Exception
    {
        checkTokenTypes(
                "<1212>",
                false,
                TokenTypes.HEX,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test name parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeName() throws Exception
    {
        checkTokenTypes(
                "/Linearized",
                false,
                TokenTypes.NAME,
                TokenTypes.END_DOCUMENT
                );
    } 

    /**
     * Test boolean parsing.
     * 
     * Ensure that null is treated as an extra event type
     * and not handled as KEYWORD.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeNull() throws Exception
    {
        checkTokenTypes(
                "null",
                false,
                TokenTypes.NULL,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test array parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeArray() throws Exception
    {
        checkTokenTypes(
                "[]",
                false,
                TokenTypes.ARRAY_START,
                TokenTypes.ARRAY_END,
                TokenTypes.END_DOCUMENT
                );
        checkTokenTypes(
                "[12]",
                false,
                TokenTypes.ARRAY_START,
                TokenTypes.NUMBER,
                TokenTypes.ARRAY_END,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test dictionary parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeDictionary() throws Exception
    {
        checkTokenTypes(
                "<<>>",
                false,
                TokenTypes.DICTIONARY_START,
                TokenTypes.DICTIONARY_END,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test stream parsing.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void isTypeStream() throws Exception
    {
        checkTokenTypes(
                "stream\r\n1234567890\rendstream",
                false,
                TokenTypes.STREAM_START,
                TokenTypes.STREAM,
                TokenTypes.STREAM_END,
                TokenTypes.END_DOCUMENT
                );
    }

    /*
     * Tests to check some transitions with similar type delimiters
     */
    
    /**
     * Test transition between Hex and Dictionary as the share
     * the same delimiter.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void testHexDictionaryTransition() throws Exception
    {
        // Hex string just after the dictionary delimiters
        checkTokenTypes(
                "<<<1212>>>",
                false,
                TokenTypes.DICTIONARY_START,
                TokenTypes.HEX,
                TokenTypes.DICTIONARY_END,
                TokenTypes.END_DOCUMENT
                );
        // hex following a dictionary
        checkTokenTypes(
                "<<12>><1212>",
                false,
                TokenTypes.DICTIONARY_START,
                TokenTypes.NUMBER,
                TokenTypes.DICTIONARY_END,
                TokenTypes.HEX,
                TokenTypes.END_DOCUMENT
                );
        // dictionary following a hex
        checkTokenTypes(
                "<1212><<12>>",
                false,
                TokenTypes.HEX,
                TokenTypes.DICTIONARY_START,
                TokenTypes.NUMBER,
                TokenTypes.DICTIONARY_END,
                TokenTypes.END_DOCUMENT
                );
    }

    /**
     * Test parsing an array within a dictionary.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void hasArrayInDictionary() throws Exception
    {
        checkTokenTypes(
                "<<[12]>>",
                false,
                TokenTypes.DICTIONARY_START,
                TokenTypes.ARRAY_START,
                TokenTypes.NUMBER,
                TokenTypes.ARRAY_END,
                TokenTypes.DICTIONARY_END,
                TokenTypes.END_DOCUMENT
                );
    }
    
    /**
     * Test recognizing delimiters correctly regardless of
     * whitespace in between or not.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void handleDelimiters() throws Exception
    {
        checkTokenTypes(
                "<</Size/Root/Info/ID>>",
                false,
                TokenTypes.DICTIONARY_START,
                TokenTypes.NAME,
                TokenTypes.NAME,
                TokenTypes.NAME,
                TokenTypes.NAME,
                TokenTypes.DICTIONARY_END,
                TokenTypes.END_DOCUMENT
                );

        checkTokenTypes(
                "<< /Size /Root /Info /ID >>",
                true,
                TokenTypes.DICTIONARY_START,
                TokenTypes.NAME,
                TokenTypes.NAME,
                TokenTypes.NAME,
                TokenTypes.NAME,
                TokenTypes.DICTIONARY_END,
                TokenTypes.END_DOCUMENT
                );
    }
    
    

    /*
     * Some tests from real world PDFs
     */
    
    
    /**
     * Test parsing of a dictionary taken from a sample pdf.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void dictionaryLexer() throws Exception
    {
        checkTokenTypes(
                "<</Size 11/Root 7 0 R/Info 5 0 R/ID" +
                "[<0E07F9F1B94B4032B20108E0BE6D152A><ECF7B96420E74436AD21C1A32A71B655>]/Prev 3629>>",
                true,
                TokenTypes.DICTIONARY_START,
                TokenTypes.NAME,
                TokenTypes.NUMBER,
                TokenTypes.NAME,
                TokenTypes.NUMBER,
                TokenTypes.NUMBER,
                TokenTypes.KEYWORD,
                TokenTypes.NAME,
                TokenTypes.NUMBER,
                TokenTypes.NUMBER,
                TokenTypes.KEYWORD,
                TokenTypes.NAME,
                TokenTypes.ARRAY_START,
                TokenTypes.HEX,
                TokenTypes.HEX,
                TokenTypes.ARRAY_END,
                TokenTypes.NAME,
                TokenTypes.NUMBER,
                TokenTypes.DICTIONARY_END,
                TokenTypes.END_DOCUMENT
                );
    }


    /**
     * Test repositioning in stream.
     * 
     * @throws Exception in case of a test failure.
     */
    @Test
    public void repositionTest() throws Exception
    {
        String content = "<</Size 11/Root 7 0 R/Info 5 0 R/ID" +
                "[<0E07F9F1B94B4032B20108E0BE6D152A><ECF7B96420E74436AD21C1A32A71B655>]/Prev 3629>>";
        RandomAccessRead source = new RandomAccessByteArray(content.getBytes());

        PDFLexer lexer = new PDFLexer(source);
        lexer.setSkipWhitespace(true);

        Assert.assertEquals(TokenTypes.BEGIN_DOCUMENT, lexer.getTokenType());
        lexer.next();lexer.next();
        Assert.assertEquals(TokenTypes.NAME, lexer.getTokenType());
        lexer.seek(0);
        Assert.assertEquals(TokenTypes.BEGIN_DOCUMENT, lexer.getTokenType());
        lexer.seek(2);
        Assert.assertEquals(TokenTypes.UNKNOWN, lexer.getTokenType());
        lexer.next();
        Assert.assertEquals(TokenTypes.NAME, lexer.getTokenType());

        // reposition into whitespace being ignored
        lexer.seek(7); lexer.next();
        Assert.assertEquals(TokenTypes.NUMBER, lexer.getTokenType());

        // reposition as above but with whitespace not ignored
        lexer.setSkipWhitespace(false);
        lexer.seek(7); lexer.next();
        Assert.assertEquals(TokenTypes.WHITESPACE, lexer.getTokenType());
    }
}